from ..const import (
    TEMP_ERROR,
    TEMP_MINIMUM,
    TEMP_MAXIMUM,
    TEMP_OFF,
    TEMP_HW_OFF,
    TEMP_HW_ON,
    MAX_BOOST_INCREASE,
    WiserUnitsEnum,
)


class _WiserTemperatureFunctions(object):
    # -----------------------------------------------------------
    # Support Functions
    # -----------------------------------------------------------
    @staticmethod
    def _to_wiser_temp(
        temp: float,
        type: str = "set_heating",
        units: WiserUnitsEnum = WiserUnitsEnum.metric,
    ) -> int:
        """
        Converts from degrees C to wiser hub format
        param temp: The temperature to convert
        param type: Can be heating (default), hotwater or delta
        return: Integer
        """
        temp = int(_WiserTemperatureFunctions._validate_temperature(temp, type) * 10)

        # Convert to metric if imperial units set
        if units == WiserUnitsEnum.imperial:
            temp = _WiserTemperatureFunctions._convert_from_F(temp)

        return temp

    @staticmethod
    def _from_wiser_temp(
        temp: int,
        type: str = "set_heating",
        units: WiserUnitsEnum = WiserUnitsEnum.metric,
    ) -> float:
        """
        Converts from wiser hub format to degrees C
        param temp: The wiser temperature to convert
        return: Float
        """
        if temp is not None:
            if temp >= TEMP_ERROR:  # Fix high value from hub when lost sight of iTRV
                temp = TEMP_MINIMUM
            else:
                temp = _WiserTemperatureFunctions._validate_temperature(
                    round(temp / 10, 1), type
                )

            # Convert to imperial if imperial units set
            if units == WiserUnitsEnum.imperial:
                temp = _WiserTemperatureFunctions._convert_to_F(temp)

            return temp
        return None

    @staticmethod
    def _validate_temperature(temp: float, type: str = "set_heating") -> float:
        """
        Validates temperature value is in range of Wiser Hub allowed values
        Sets to min or max temp if value exceeds limits
        param temp: temperature value to validate
        return: float
        """

        # Accomodate hw temps
        if type == "hotwater" and temp in [TEMP_HW_ON, TEMP_HW_OFF]:
            return temp

        # Accomodate temp deltas
        if type == "delta":
            if temp > MAX_BOOST_INCREASE:
                return MAX_BOOST_INCREASE
            return temp

        # Accomodate reported current temps
        if type == "current":
            if temp < TEMP_OFF:
                return TEMP_MINIMUM
            return temp

        # Accomodate heating temps
        if type == "set_heating":
            if temp >= TEMP_ERROR:
                return TEMP_MINIMUM
            elif temp > TEMP_MAXIMUM:
                return TEMP_MAXIMUM
            elif temp < TEMP_MINIMUM and temp != TEMP_OFF:
                return TEMP_MINIMUM
            else:
                return temp

    @staticmethod
    def _convert_from_F(temp: float) -> float:
        """
        Convert F temp to C
        param temp: temp in F to convert
        return: float
        """
        return round((temp - 32) * 5 / 9, 1)

    @staticmethod
    def _convert_to_F(temp: float) -> float:
        """
        Convert C temp to F
        param temp: temp in C to convert
        return: float
        """
        return round((temp * 9 / 5) + 32, 1)



# Exception Handlers
class WiserHubConnectionError(Exception):
    pass


class WiserHubAuthenticationError(Exception):
    pass


class WiserHubRESTError(Exception):
    pass


class WiserHubNotImplementedError(Exception):
    # _LOGGER.info("Function not yet implemented")
    pass



class _WiserHubCapabilitiesInfo:
    """Data structure for capabilities info for Wiser Hub"""

    def __init__(self, data: dict):
        self._data = data

    @property
    def all(self) -> dict:
        "Get the list of capabilities"
        return dict(self._data)

    @property
    def smartplug(self):
        return self._data.get("SmartPlug", False)

    @property
    def itrv(self):
        return self._data.get("ITRV", False)

    @property
    def roomstat(self):
        return self._data.get("Roomstat", False)

    @property
    def ufh(self):
        return self._data.get("UFH", False)

    @property
    def ufh_floor_temp_sensor(self):
        return self._data.get("UFHFloorTempSensor", False)

    @property
    def ufh_dew_sensor(self):
        return self._data.get("UFHDewSensor", False)

    @property
    def hact(self):
        return self._data.get("HACT", False)

    @property
    def lact(self):
        return self._data.get("LACT", False)

    @property
    def light(self):
        return self._data.get("Light", False)

    @property
    def shutter(self):
        return self._data.get("Shutter", False)

    @property
    def load_controller(self):
        return self._data.get("LoadController", False)



#!/usr/bin/env python3
"""
# Wiser API Version 2

angelosantagata@gmail.com
msparker@sky.com


https://github.com/asantaga/wiserheatingapi


This API allows you to get information from and control your wiserhub.
"""

# TODO: Keep objects and update instead of recreating on hub update
# TODO: Update entity values after commend issued to get current values
import asyncio
import aiohttp
import pathlib
from typing import Optional
from . import _LOGGER, __VERSION__

from .const import (
    DEFAULT_AWAY_MODE_TEMP,
    DEFAULT_DEGRADED_TEMP,
    MAX_BOOST_INCREASE,
    TEMP_ERROR,
    TEMP_HW_ON,
    TEMP_HW_OFF,
    TEMP_MINIMUM,
    TEMP_MAXIMUM,
    TEMP_OFF,
    WiserUnitsEnum,
    WISERHUBDOMAIN,
    WISERHUBNETWORK,
    WISERHUBSCHEDULES,
    WISERHUBOPENTHERM,
)

from .exceptions import (
    WiserHubConnectionError,
    WiserHubAuthenticationError,
    WiserHubRESTError,
)

from .cli import log_response_to_file
from .devices import _WiserDeviceCollection
from .heating import _WiserHeatingChannelCollection
from .hot_water import _WiserHotwater
from .moments import _WiserMomentCollection
from .rest_controller import _WiserRestController, _WiserConnectionInfo
from .room import _WiserRoomCollection
from .schedule import _WiserScheduleCollection, WiserScheduleTypeEnum
from .system import _WiserSystem


class WiserAPI(object):
    """
    Main api class to access all entities and attributes of wiser system
    """

    def __init__(
        self,
        host: str,
        secret: str,
        port: Optional[int] = 80,
        session: Optional[aiohttp.ClientSession] = None,
        units: Optional[WiserUnitsEnum] = WiserUnitsEnum.metric,
    ):

        # Connection variables
        self._session = session
        self._wiser_api_connection = _WiserConnectionInfo()
        self._wiser_rest_controller = None

        # Set connection params
        self._wiser_api_connection.host = host
        self._wiser_api_connection.secret = secret
        self._wiser_api_connection.port = port
        self._wiser_api_connection.units = units

        # Hub Data
        self._domain_data = {}
        self._network_data = {}
        self._schedule_data = {}
        self._opentherm_data = {}

        # Data stores for exposed properties
        self._devices = None
        self._hotwater = None
        self._heating_channels = None
        self._moments = None
        self._rooms = None
        self._schedules = None
        self._system = None

        # Log initialisation info
        _LOGGER.info(
            f"WiserHub API v{__VERSION__} Initialised - Host: {host}, Units: {self._wiser_api_connection.units.name.title()}"
        )

        if (
            self._wiser_api_connection.host is not None
            and self._wiser_api_connection.secret is not None
        ):
            # Create an instance of the rest controller
            self._wiser_rest_controller = _WiserRestController(
                self._session, wiser_connection_info=self._wiser_api_connection
            )
        else:
            raise WiserHubConnectionError(
                "Missing or incomplete connection information"
            )

    async def read_hub_data(self):
        """Read all data from hub and populate objects"""

        # Read data from hub
        self._domain_data = await self._wiser_rest_controller._get_hub_data(
            WISERHUBDOMAIN
        )
        self._network_data = await self._wiser_rest_controller._get_hub_data(
            WISERHUBNETWORK
        )
        self._schedule_data = await self._wiser_rest_controller._get_hub_data(
            WISERHUBSCHEDULES
        )

        # Only get opentherm data if connected
        if (
            self._domain_data.get("System", {}).get("OpenThermConnectionStatus", "")
            == "Connected"
        ):
            self._opentherm_data = await self._wiser_rest_controller._get_hub_data(
                WISERHUBOPENTHERM, False
            )

        if self._domain_data != {} and self._network_data != {}:

            # System Object
            _device_data = self._domain_data.get("Device", [])
            self._system = _WiserSystem(
                self._wiser_rest_controller,
                self._domain_data,
                self._network_data,
                _device_data,
                self._opentherm_data,
            )

            # Schedules Collection
            self._schedules = _WiserScheduleCollection(
                self._wiser_rest_controller,
                self._schedule_data,
                self._system.sunrise_times,
                self._system.sunset_times,
            )

            # Devices Collection
            self._devices = _WiserDeviceCollection(
                self._wiser_rest_controller, self._domain_data, self._schedules
            )

            # Rooms Collection
            room_data = self._domain_data.get("Room", [])
            self._rooms = _WiserRoomCollection(
                self._wiser_rest_controller,
                room_data,
                self._schedules.get_by_type(WiserScheduleTypeEnum.heating),
                self._devices,
            )

            # Hot Water
            if self._domain_data.get("HotWater"):
                schedule = self._schedules.get_by_id(
                    WiserScheduleTypeEnum.onoff,
                    self._domain_data.get("HotWater")[0].get("ScheduleId", 0),
                )
                self._hotwater = _WiserHotwater(
                    self._wiser_rest_controller,
                    self._domain_data.get("HotWater", {})[0],
                    schedule,
                )

            # Heating Channels
            if self._domain_data.get("HeatingChannel"):
                self._heating_channels = _WiserHeatingChannelCollection(
                    self._domain_data.get("HeatingChannel"), self._rooms
                )

            # Moments
            if self._domain_data.get("Moment"):
                self._moments = _WiserMomentCollection(
                    self._wiser_rest_controller, self._domain_data.get("Moment")
                )

            # If gets here with no exceptions then success and return true
            return True

        return False

    async def close_session(self):
        await self._wiser_rest_controller._session.close()

    # API properties
    @property
    def devices(self):
        """List of device entities attached to the Wiser Hub"""
        return self._devices

    @property
    def heating_channels(self):
        """List of heating channel entities on the Wiser Hub"""
        return self._heating_channels

    @property
    def hotwater(self):
        """List of hot water entities on the Wiser Hub"""
        return self._hotwater

    @property
    def moments(self):
        """List of moment entities on the Wiser Hub"""
        return self._moments

    @property
    def rooms(self):
        """List of room entities configured on the Wiser Hub"""
        return self._rooms

    @property
    def schedules(self):
        """List of schedules"""
        return self._schedules

    @property
    def system(self):
        """Entity of the Wiser Hub"""
        return self._system

    @property
    def units(self) -> WiserUnitsEnum:
        """Get or set units for temperature"""
        return self._wiser_api_connection.units

    @units.setter
    def units(self, units: WiserUnitsEnum):
        self._wiser_api_connection.units = units

    @property
    def version(self):
        return __VERSION__

    @property
    def _raw_hub_data(self):
        return {
            "Domain": self._domain_data,
            "Network": self._network_data,
            "Schedule": self._schedule_data,
            "OpenTherm": self._opentherm_data,
        }

    def output_raw_hub_data(
        self, data_class: str, filename: str, file_path: str
    ) -> bool:
        """Output raw hub data to json file"""
        # Get correct endpoint
        if data_class.lower() == "domain":
            endpoint = WISERHUBDOMAIN
        elif data_class.lower() == "network":
            endpoint = WISERHUBNETWORK
        elif data_class == "schedules":
            endpoint = WISERHUBSCHEDULES
        else:
            endpoint = None

        # Get raw json data
        if endpoint:
            data = self._wiser_rest_controller._get_hub_data(endpoint)
            try:
                if data:
                    # Write out to file
                    log_response_to_file(data, filename, False, pathlib.Path(file_path))
                    return True
            except Exception as ex:
                _LOGGER.error(ex)
                return False



from . import _LOGGER

from .const import (
    REST_TIMEOUT,
    WISERHUBDOMAIN,
    WISERHUBSCHEDULES,
    WiserUnitsEnum,
)

from .exceptions import (
    WiserHubAuthenticationError,
    WiserHubConnectionError,
    WiserHubRESTError,
)

import asyncio
import aiohttp
import enum

from typing import Any, Optional, cast

# Connection info class
class _WiserConnectionInfo(object):
    def __init(self):
        self.host = None
        self.secret = None
        self.port = None
        self.units = WiserUnitsEnum.metric


# Enums
class WiserRestActionEnum(enum.Enum):
    GET = "get"
    POST = "post"
    PATCH = "patch"
    DELETE = "delete"


class _WiserRestController(object):
    """
    Class to handle getting data from and sending commands to a wiser hub
    """

    def __init__(
        self,
        session: Optional[aiohttp.ClientSession] = None,
        timeout: Optional[float] = REST_TIMEOUT,
        wiser_connection_info: Optional[_WiserConnectionInfo] = None,
    ):

        self._wiser_connection_info = wiser_connection_info

        if not session:
            session = aiohttp.ClientSession()
        self._session = session
        self._timeout = aiohttp.ClientTimeout(total=timeout)

    async def _do_hub_action(
        self,
        action: WiserRestActionEnum,
        url: str,
        data: dict = None,
        raise_for_endpoint_error: bool = True,
    ):
        """
        Send patch update to hub and raise errors if fails
        param url: url of hub rest api endpoint
        param patchData: json object containing command and values to set
        return: boolean
        """

        kwargs = {
            "headers": {
                "SECRET": self._wiser_connection_info.secret,
                "Content-Type": "application/json;charset=UTF-8",
            }
        }

        if data is not None:
            # print(data)
            kwargs["json"] = data
        if self._timeout is not None:
            kwargs["timeout"] = self._timeout

        try:
            response = cast(
                aiohttp.ClientResponse,
                await getattr(self._session, action.value)(
                    url.format(
                        self._wiser_connection_info.host,
                        self._wiser_connection_info.port,
                    ),
                    **kwargs,
                ),
            )

            if not response.ok:
                self._process_nok_response(response, raise_for_endpoint_error)
            else:
                #    if action == WiserRestActionEnum.GET:
                if len(await response.text()) > 0:
                    # response = re.sub(rb'[^\x20-\x7F]+', b'', response.content.iter_any())
                    return await response.json()
                else:
                    return {}
            return {}

        except asyncio.TimeoutError as ex:
            raise WiserHubConnectionError(
                f"Connection timeout trying to communicate with Wiser Hub {self._wiser_connection_info.host}.  Error is {ex}"
            )
        except aiohttp.ClientResponseError as ex:
            raise WiserHubConnectionError(
                f"Response error trying to communicate with Wiser Hub {self._wiser_connection_info.host}.  Error is {ex}"
            )
        except aiohttp.ClientConnectorError as ex:
            raise WiserHubConnectionError(
                f"Connection error trying to communicate with Wiser Hub {self._wiser_connection_info.host}.  Error is {ex}"
            )

    def _process_nok_response(
        self, response: aiohttp.ClientResponse, raise_for_endpoint_error: bool = True
    ):
        if response.status == 401:
            raise WiserHubAuthenticationError(
                f"Error authenticating to Wiser Hub {self._wiser_connection_info.host}.  Check your secret key"
            )
        elif response.status == 404 and raise_for_endpoint_error:
            raise WiserHubRESTError(
                f"Rest endpoint not found on Wiser Hub {self._wiser_connection_info.host}"
            )
        elif response.status == 408:
            raise WiserHubConnectionError(
                f"Connection timed out trying to communicate with Wiser Hub {self._wiser_connection_info.host}"
            )
        elif raise_for_endpoint_error:
            self._session.close()
            raise WiserHubRESTError(
                f"Unknown error getting communicating with Wiser Hub {self._wiser_connection_info.host}.  Error code is: {response.status}"
            )
        else:
            print("Error of the Unknown kind!")

    async def _get_hub_data(self, url: str, raise_for_endpoint_error: bool = True):
        """Get data from hub"""
        return await self._do_hub_action(
            WiserRestActionEnum.GET,
            url,
            raise_for_endpoint_error=raise_for_endpoint_error,
        )

    async def _send_command(
        self,
        url: str,
        command_data: dict,
        method: WiserRestActionEnum = WiserRestActionEnum.PATCH,
    ):
        """
        Send control command to hub and raise errors if fails
        param url: url of hub rest api endpoint
        param patchData: json object containing command and values to set
        return: boolean
        """
        url = WISERHUBDOMAIN + url
        _LOGGER.debug(
            "Sending command to url: {} with parameters {}".format(url, command_data)
        )

        return await self._do_hub_action(method, url, command_data)

    async def _do_schedule_action(
        self, action: WiserRestActionEnum, url: str, schedule_data: dict = None
    ):
        """
        Perform schedule action to hub and raise errors if fails
        param url: url of hub rest api endpoint
        param patchData: json object containing schedule values to set
        return: boolean
        """
        url = WISERHUBSCHEDULES + url
        _LOGGER.debug(
            "Actioning schedule to url: {} with action {} and data {}".format(
                url, action.value, schedule_data
            )
        )
        return await self._do_hub_action(action, url, schedule_data)

    async def _send_schedule_command(
        self, action: str, schedule_data: dict, id: int = 0, schedule_type: str = None
    ) -> bool:
        """
        Send schedule data to Wiser Hub
        param schedule_data: json schedule data
        param id: schedule id
        return: boolen - true = success, false = failed
        """
        if action == "UPDATE":
            result = await self._do_schedule_action(
                WiserRestActionEnum.PATCH,
                "{}/{}".format(schedule_type, id),
                schedule_data,
            )

        elif action == "CREATE":
            result = await self._do_schedule_action(
                WiserRestActionEnum.POST,
                "Assign",
                schedule_data,
            )

        elif action == "ASSIGN":
            result = await self._do_schedule_action(
                WiserRestActionEnum.PATCH,
                "Assign",
                schedule_data,
            )

        elif action == "DELETE":
            result = await self._do_schedule_action(
                WiserRestActionEnum.DELETE,
                "{}/{}".format(schedule_type, id),
                schedule_data,
            )
        return result



from . import _LOGGER

from .const import TEMP_OFF, TEXT_UNKNOWN, WISERHEATINGACTUATOR, WISERDEVICE
from .helpers.device import _WiserDevice
from .helpers.temp import _WiserTemperatureFunctions as tf


class _WiserHeatingActuator(_WiserDevice):
    """Class representing a Wiser Heating Actuator device"""

    @property
    def current_target_temperature(self) -> float:
        """Get the smart valve current target temperature setting"""
        return tf._from_wiser_temp(
            self._device_type_data.get("OccupiedHeatingSetPoint", TEMP_OFF)
        )

    @property
    def current_temperature(self) -> float:
        """Get the current temperature measured by the smart valve"""
        return tf._from_wiser_temp(
            self._device_type_data.get("MeasuredTemperature", TEMP_OFF), "current"
        )

    @property
    def delivered_power(self) -> int:
        """Get the amount of current throught the plug over time"""
        return self._device_type_data.get("CurrentSummationDelivered", 0)

    @property
    def instantaneous_power(self) -> int:
        """Get the amount of current throught the plug now"""
        return self._device_type_data.get("InstantaneousDemand", 0)

    @property
    def output_type(self) -> str:
        """Get output type"""
        return self._device_type_data.get("OutputType", TEXT_UNKNOWN)


class _WiserHeatingActuatorCollection(object):
    """Class holding all wiser heating actuators"""

    def __init__(self):
        self._items = []

    @property
    def all(self) -> dict:
        return list(self._items)

    @property
    def count(self) -> int:
        return len(self.all)

    def get_by_id(self, id: int) -> _WiserHeatingActuator:
        """
        Gets a Heating Actuator object from the Heating Actuators id
        param id: id of smart valve
        return: _WiserSmartValve object
        """
        try:
            return [
                heating_actuator
                for heating_actuator in self.all
                if heating_actuator.id == id
            ][0]
        except IndexError:
            return None
